#!/usr/bin/env python
# -*- encoding: utf-8 -*-
"""
@File    :  tidalImp.py
@Date    :  2021/9/2
@Author  :  Yaronzz
@Version :  1.0
@Contact :  yaronhuang@foxmail.com
@Desc    :
"""
import time

import requests
import tidal_dl.model
import tidal_dl.enums
from aigpy.stringHelper import isNull
from tidal_dl import TokenSettings, TOKEN, TidalAPI, CONF
from tidal_dl.util import getAlbumPath, getPlaylistPath, getTrackPath


class TidalImp(TidalAPI):
    def __init__(self):
        super(TidalImp, self).__init__()

    def loginByConfig(self):
        if isNull(TOKEN.accessToken):
            return False

        msg, check = self.verifyAccessToken(TOKEN.accessToken)
        if check:
            self.key.countryCode = TOKEN.countryCode
            self.key.userId = TOKEN.userid
            self.key.accessToken = TOKEN.accessToken
            return True

        msg, check = self.refreshAccessToken(TOKEN.refreshToken)
        if check:
            TOKEN.userid = self.key.userId
            TOKEN.countryCode = self.key.countryCode
            TOKEN.accessToken = self.key.accessToken
            TOKEN.expiresAfter = time.time() + int(self.key.expiresIn)
            TokenSettings.save(TOKEN)
            return True
        else:
            tmp = TokenSettings()  # clears saved tokens
            TokenSettings.save(tmp)
            return False

    def loginByWeb(self):
        start = time.time()
        elapsed = 0
        while elapsed < self.key.authCheckTimeout:
            elapsed = time.time() - start
            msg, check = self.checkAuthStatus()
            if not check:
                if msg == "pending":
                    time.sleep(self.key.authCheckInterval + 1)
                    continue
                return False
            if check:
                TOKEN.userid = self.key.userId
                TOKEN.countryCode = self.key.countryCode
                TOKEN.accessToken = self.key.accessToken
                TOKEN.refreshToken = self.key.refreshToken
                TOKEN.expiresAfter = time.time() + int(self.key.expiresIn)
                TokenSettings.save(TOKEN)
                return True
        return False

    @staticmethod
    def getArtistsNames(artists):  # : list[tidal_dl.model.Artist]
        ret = []
        for item in artists:
            ret.append(item.name)
        return ','.join(ret)

    @staticmethod
    def getDurationString(seconds: int):
        m, s = divmod(seconds, 60)
        h, m = divmod(m, 60)
        return "%02d:%02d:%02d" % (h, m, s)

    def getCoverData(self, sid, width="320", height="320"):
        url = self.getCoverUrl(sid, width, height)
        try:
            respond = requests.get(url)
            return respond.content
        except:
            return ''
    
    @staticmethod 
    def getAudioQualityList():
        return map(lambda quality: quality.name, tidal_dl.enums.AudioQuality)
    
    @staticmethod
    def getVideoQualityList():
        return map(lambda quality: quality.name, tidal_dl.enums.VideoQuality)
    
    def getBasePath(self, model):
        if isinstance(model, tidal_dl.model.Album):
            return getAlbumPath(CONF, model)
        if isinstance(model, tidal_dl.model.Playlist):
            return getPlaylistPath(CONF, model)
        if isinstance(model, tidal_dl.model.Track):
            return getAlbumPath(CONF, model.album)
        if isinstance(model, tidal_dl.model.Video):
            return CONF.downloadPath + '/Video/'
        return './'

    def getConfig(self):
        return CONF
    
    # def getTackPath(self, basePath, track, stream, album=None, playlist=None):
    #     # number
    #     number = __getIndexStr__(track.trackNumber)
    #     if playlist is not None and CONF.usePlaylistFolder:
    #         number = __getIndexStr__(track.trackNumberOnPlaylist)
    #     # artist
    #     artist = aigpy.path.replaceLimitChar(__getArtistsString__(track.artists), '-')
    #     # title
    #     title = track.title
    #     if not aigpy.string.isNull(track.version):
    #         title += ' (' + track.version + ')'
    #     title = aigpy.path.replaceLimitChar(title, '-')
    #     # get explicit
    #     explicit = "(Explicit)" if CONF.addExplicitTag and track.explicit else ''
    #     # album and addyear
    #     albumname = aigpy.path.replaceLimitChar(album.title, '-')
    #     year = ""
    #     if album.releaseDate is not None:
    #         year = aigpy.string.getSubOnlyEnd(album.releaseDate, '-')
    #     # extension
    #     extension = __getExtension__(stream.url)
    #     retpath = CONF.trackFileFormat
    #     if retpath is None or len(retpath) <= 0:
    #         retpath = Settings.getDefaultTrackFileFormat()
    #     retpath = retpath.replace(R"{TrackNumber}", number)
    #     retpath = retpath.replace(R"{ArtistName}", artist.strip())
    #     retpath = retpath.replace(R"{TrackTitle}", title)
    #     retpath = retpath.replace(R"{ExplicitFlag}", explicit)
    #     retpath = retpath.replace(R"{AlbumYear}", year)
    #     retpath = retpath.replace(R"{AlbumTitle}", albumname.strip())
    #     retpath = retpath.strip()
    #     return basePath + retpath + extension
    
    # def getVideoPath(self, basePath, video):
    #     # hyphen
    #     hyphen = ' - ' if CONF.addHyphen else ' '
    #     # get number
    #     number = ''
    #     if CONF.useTrackNumber:
    #         number = __getIndexStr__(video.trackNumber) + hyphen
    #     # get artist
    #     artist = ''
    #     if CONF.artistBeforeTitle:
    #         artist = aigpy.path.replaceLimitChar(__getArtistsString__(video.artists), '-') + hyphen
    #     # get explicit
    #     explicit = "(Explicit)" if CONF.addExplicitTag and video.explicit else ''
    #     # title
    #     title = aigpy.path.replaceLimitChar(video.title, '-')
    #     # extension
    #     extension = ".mp4"
    #     return basePath + number + artist.strip() + title + explicit + extension
        
tidalImp = TidalImp()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
"""
@File    :  downloadItemModel.py
@Date    :  2021/10/08
@Author  :  Yaronzz
@Version :  1.0
@Contact :  yaronhuang@foxmail.com
@Desc    :  
"""

import _thread
import os
import time
from enum import Enum
import aigpy.stringHelper
from tidal_dl import Type
from tidal_dl.model import Album, Track, Video, Playlist

from tidal_gui.tidalImp import tidalImp
from tidal_gui.view.downloadItemView import DownloadItemView
from tidal_gui.viewModel.viewModel import ViewModel

class DownloadStatus(Enum):
    Wait = 0,
    Running = 1,
    Finish = 2,
    Error = 3,
    Cancel = 4,
    
_endStatus_ = [DownloadStatus.Finish, DownloadStatus.Error, DownloadStatus.Cancel]

class DownloadItemModel(ViewModel):
    def __init__(self, index, data, basePath):
        super(DownloadItemModel, self).__init__()
        self.view = DownloadItemView()
        self.data = data
        self.basePath = basePath
        self.isTrack = isinstance(data, Track)
        self.__setStatus__(DownloadStatus.Wait)
        
        if self.isTrack:
            self.__initTrack__(index)
        else:
            self.__initVideo__(index)

    def __setStatus__(self, status: DownloadStatus, desc: str = ''):
        self.status = status
        if desc == '':
            self.view.setAction(status.name)
        else:
            self.view.setAction(status.name + '-' + desc)
    
    def __initTrack__(self, index):
        title = self.data.title
        own = self.data.album.title
        self.view.setLabel(index, title, own)

    def __initVideo__(self, index):
        title = self.data.title
        own = tidalImp.getArtistsNames(self.data.artists)
        self.view.setLabel(index, title, own)

    def isInWait(self):
        return self.status == DownloadStatus.Wait
        
    def stopDownload(self):
        if self.status not in _endStatus_:
            self.__setStatus__(DownloadStatus.Cancel)

    def retry(self):
        self.__setStatus__(DownloadStatus.Wait)
        
    def download(self):
        self.__setStatus__(DownloadStatus.Running)
        
        if self.isTrack:
            if not self.__dlTrack__():
                return
        else:
            if not self.__dlVideo__():
                return
            
        self.view.setProgress(100)
        self.__setStatus__(DownloadStatus.Finish)
    
    def __dlTrack__(self):
        try:
            track = self.data
            conf = tidalImp.getConfig()
            
            msg, stream = tidalImp.getStreamUrl(track.id, conf.audioQuality)
            if not aigpy.string.isNull(msg) or stream is None:
                self.__setStatus__(DownloadStatus.Error)
                return False
            
            tidalImp.getBasePath(track)
            path = tidalImp.getTackPath(self.basePath, track, stream) 

            # # check exist
            # if conf.checkExist and tidalImp.__isNeedDownload__(path, stream.url) == False:
            #     return True
            
            tool = aigpy.download.DownloadTool(path + '.part', [stream.url])
            check, err = tool.start(conf.showProgress)

            if not check:
                Printf.err("Download failed! " + aigpy.path.getFileName(path) + ' (' + str(err) + ')')
                return
            # encrypted -> decrypt and remove encrypted file
            if aigpy.string.isNull(stream.encryptionKey):
                os.replace(path + '.part', path)
            else:
                key, nonce = decrypt_security_token(stream.encryptionKey)
                decrypt_file(path + '.part', path, key, nonce)
                os.remove(path + '.part')

            path = __convertToM4a__(path, stream.codec)

            # contributors
            msg, contributors = API.getTrackContributors(track.id)
            msg, tidalLyrics = API.getLyrics(track.id)

            lyrics = '' if tidalLyrics is None else tidalLyrics.subtitles
            if conf.addLyrics and lyrics == '':
                lyrics = __getLyrics__(track.title, __getArtistsString__(track.artists), conf.lyricsServerProxy)

            if conf.lyricFile:
                if tidalLyrics is None:
                    Printf.info(f'Failed to get lyrics from tidal!"{track.title}"')
                else:
                    lrcPath = path.rsplit(".", 1)[0] + '.lrc'
                    aigpy.fileHelper.write(lrcPath, tidalLyrics.subtitles, 'w')

            __setMetaData__(track, album, path, contributors, lyrics)
            Printf.success(aigpy.path.getFileName(path))
        except Exception as e:
            Printf.err("Download failed! " + track.title + ' (' + str(e) + ')')
    
    def __dlVideo__(self):
        pass



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
"""
@File    :  framelessWidget.py
@Date    :  2021/05/08
@Author  :  Yaronzz
@Version :  1.0
@Contact :  yaronhuang@foxmail.com
@Desc    :
"""
import ctypes.wintypes
from ctypes.wintypes import POINT
import typing

import PyQt5
import win32api
import win32con
from PyQt5.QtCore import Qt, QPoint
from PyQt5.QtGui import QMouseEvent
from PyQt5.QtWidgets import QWidget, QGridLayout, QHBoxLayout

from tidal_gui.control.pushButton import PushButton
from tidal_gui.style import ButtonStyle


class FramelessWidget(QWidget):
    def __init__(self):
        super(FramelessWidget, self).__init__()
        self.setWindowFlags(Qt.FramelessWindowHint)
        self.BorderWidth = 5
        self.borderWidget = QWidget()
        self.borderWidget.setObjectName("widgetMain")
        self.borderWidget.setStyleSheet("QWidget#widgetMain{border: 1px solid #000000;};")

        self.contentGrid = QGridLayout()
        self.contentGrid.setContentsMargins(1, 1, 1, 1)

        self.windowBtnGrid = self.__createWindowsButtonLayout__()
        self.enableMove = True
        self.validMoveWidget = None
        self.clickPos = None

        self.grid = QGridLayout()
        self.grid.setSpacing(0)
        self.grid.setContentsMargins(0, 0, 0, 0)
        self.grid.addWidget(self.borderWidget, 0, 0)
        self.grid.addLayout(self.contentGrid, 0, 0)
        self.setLayout(self.grid)

    def __showMaxWindows__(self):
        if self.windowState() == Qt.WindowMaximized:
            self.showNormal()
        else:
            self.showMaximized()

    def __createWindowsButtonLayout__(self):
        self.closeBtn = PushButton('', ButtonStyle.CloseWindow)
        self.maxBtn = PushButton('', ButtonStyle.MaxWindow)
        self.minBtn = PushButton('', ButtonStyle.MinWindow)

        self.closeBtn.clicked.connect(self.close)
        self.minBtn.clicked.connect(self.showMinimized)
        self.maxBtn.clicked.connect(self.__showMaxWindows__)

        layout = QHBoxLayout()
        layout.setSpacing(0)
        layout.setContentsMargins(1, 1, 1, 1)
        layout.addWidget(self.minBtn)
        layout.addWidget(self.maxBtn)
        layout.addWidget(self.closeBtn)
        return layout

    def __clickInValidMoveWidget__(self, x=-1, y=-1) -> bool:
        if self.validMoveWidget is None:
            return True
        if x == -1 and y == -1:
            x = self.clickPos.x()
            y = self.clickPos.y()

        pos = self.validMoveWidget.pos()
        if x < pos.x() or x > pos.x() + self.validMoveWidget.width():
            return False
        if y < pos.y() or y > pos.y() + self.validMoveWidget.height():
            return False
        return True

    def nativeEvent(self, eventType, message):
        retVal, result = super(FramelessWidget, self).nativeEvent(eventType, message)
        # if eventType == "windows_generic_MSG":
        #     msg = ctypes.wintypes.MSG.from_address(message.__int__())
        #     if msg.message != win32con.WM_NCHITTEST:
        #         return retVal, result

        #     # 获取鼠标移动经过时的坐标
        #     x = win32api.LOWORD(msg.lParam) - self.frameGeometry().x()
        #     y = win32api.HIWORD(msg.lParam) - self.frameGeometry().y()

        #     w, h = self.width(), self.height()
        #     lx = x < self.BorderWidth
        #     rx = x > w - self.BorderWidth
        #     ty = y < self.BorderWidth
        #     by = y > h - self.BorderWidth
        #     if (lx and ty):# 左上角
        #         return True, win32con.HTTOPLEFT
        #     elif (rx and by):# 右下角
        #         return True, win32con.HTBOTTOMRIGHT
        #     elif (rx and ty):# 右上角
        #         return True, win32con.HTTOPRIGHT
        #     elif (lx and by):# 左下角
        #         return True, win32con.HTBOTTOMLEFT
        #     elif ty:# 上
        #         return True, win32con.HTTOP
        #     elif by:# 下
        #         return True, win32con.HTBOTTOM
        #     elif lx:# 左
        #         return True, win32con.HTLEFT
        #     elif rx:# 右
        #         return True, win32con.HTRIGHT
        return retVal, result

    def mousePressEvent(self, e: QMouseEvent):
        if e.button() == Qt.LeftButton:
            self.clickPos = e.pos()

    def mouseReleaseEvent(self, e: QMouseEvent):
        if e.button() == Qt.LeftButton:
            self.clickPos = QPoint(-1, -1)

    def mouseMoveEvent(self, e: QMouseEvent):
        if not self.enableMove:
            return
        if Qt.LeftButton & e.buttons():
            if self.__clickInValidMoveWidget__() and self.clickPos:
                self.move(e.pos() + self.pos() - self.clickPos)

    def mouseDoubleClickEvent(self, e: QMouseEvent):
        if self.maxBtn.isHidden():
            return
        if Qt.LeftButton & e.buttons():
            if self.__clickInValidMoveWidget__(e.x(), e.y()):
                self.__showMaxWindows__()

    def getGrid(self):
        return self.contentGrid

    def disableMove(self):
        self.enableMove = False

    def setValidMoveWidget(self, widget):
        self.validMoveWidget = widget

    def setWindowButton(self, showClose=True, showMin=True, showMax=True):
        if not showMax:
            self.maxBtn.hide()
        if not showMin:
            self.minBtn.hide()
        if not showClose:
            self.closeBtn.hide()
        self.grid.addLayout(self.windowBtnGrid, 0, 0, Qt.AlignTop | Qt.AlignRight)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
"""
@File    :  scrollWidget.py
@Date    :  2021/10/08
@Author  :  Yaronzz
@Version :  1.0
@Contact :  yaronhuang@foxmail.com
@Desc    :  
"""

from PyQt5.QtCore import QSize, Qt
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QListWidget, QListWidgetItem, QWidget, QScrollArea, QVBoxLayout, QPushButton, QBoxLayout


class ScrollWidget(QScrollArea):
    def __init__(self):
        super(ScrollWidget, self).__init__()
        
        self._layout = QVBoxLayout()
        self._layout.addStretch(1)
        
        self._mainW = QWidget()
        self._mainW.setLayout(self._layout)
        
        self.setWidget(self._mainW)
        self.setWidgetResizable(True)

    def addWidgetItem(self, widget: QWidget):
        self._layout.insertWidget(0, widget)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
"""
@File    :  tableWidget.py
@Date    :  2021/8/18
@Author  :  Yaronzz
@Version :  1.0
@Contact :  yaronhuang@foxmail.com
@Desc    :
"""
import aigpy
import requests
import threading
from PyQt5.QtCore import Qt, QUrl
from PyQt5.QtGui import QIcon, QPixmap
from PyQt5.QtNetwork import QNetworkAccessManager, QNetworkRequest, QNetworkReply
from PyQt5.QtWidgets import QTableWidget, QTableWidgetItem, QAbstractItemView, QLabel
from tidal_gui.control.label import Label, LabelStyle


class TableWidget(QTableWidget):
    def __init__(self, columnNames: list, rowCount: int = 20):
        super(TableWidget, self).__init__()
        self.setColumnCount(len(columnNames))
        self.setRowCount(rowCount)

        self._lock = threading.Lock()
        self._netManager = QNetworkAccessManager()

        self.columnAligns = []

        for index, name in enumerate(columnNames):
            item = QTableWidgetItem(name)

            align = Qt.AlignLeft | Qt.AlignVCenter
            if name == '#' or name == ' ':
                align = Qt.AlignCenter

            self.columnAligns.append(align)
            item.setTextAlignment(align)
            self.setHorizontalHeaderItem(index, item)

        for index in range(0, rowCount):
            self.setRowHeight(index, 50)

        self.setShowGrid(False)
        self.verticalHeader().setVisible(False)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setSelectionMode(QAbstractItemView.SingleSelection)

        self.horizontalHeader().setStretchLastSection(True)
        self.setEditTriggers(QAbstractItemView.NoEditTriggers)

        self.setFocusPolicy(Qt.NoFocus)

    def changeRowCount(self, rows: int):
        if rows != self.rowCount():
            self.setRowCount(rows)
            for index in range(0, rows):
                self.setRowHeight(index, 50)

    def addItem(self, rowIdx: int, colIdx: int, text):
        if isinstance(text, str):
            item = QTableWidgetItem(text)
            item.setTextAlignment(self.columnAligns[colIdx])
            self.setItem(rowIdx, colIdx, item)
        elif isinstance(text, QUrl):
            self.__addPicItem__(rowIdx, colIdx, text)

    def __picDownload__(self, rowIdx, colIdx):
        reply = self.sender()
        data = reply.readAll()
        if data.size() <= 0:
            return

        pic = QPixmap()
        pic.loadFromData(data)

        self._lock.acquire()
        self.cellWidget(rowIdx, colIdx).setPixmap(pic.scaled(32, 32))
        self._lock.release()

        reply.deleteLater()

    def __addPicItem__(self, rowIdx: int, colIdx: int, url: QUrl):
        icon = Label('', LabelStyle.Icon)
        icon.setAlignment(Qt.AlignCenter)

        self.setCellWidget(rowIdx, colIdx, icon)

        reply = self._netManager.get(QNetworkRequest(url))
        reply.finished.connect(lambda: self.__picDownload__(rowIdx, colIdx))

    def addWidgetItem(self, rowIdx: int, colIdx: int, widget):
        self.setCellWidget(rowIdx, colIdx, widget)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
"""
@File    :  taskItemModel.py
@Date    :  2021/9/14
@Author  :  Yaronzz
@Version :  1.0
@Contact :  yaronhuang@foxmail.com
@Desc    :
"""
import _thread
import os

import aigpy.stringHelper
from tidal_dl import Type
from tidal_dl.model import Album, Track, Video, Playlist

from tidal_gui.tidalImp import tidalImp
from tidal_gui.view.taskItemView import TaskItemView
from tidal_gui.viewModel.viewModel import ViewModel
from tidal_gui.viewModel.downloadItemModel import DownloadItemModel
from PyQt5.QtWidgets import QListWidget, QListWidgetItem, QWidget, QScrollArea, QVBoxLayout, QPushButton
import time


class TaskItemModel(ViewModel):
    def __init__(self, data):
        super(TaskItemModel, self).__init__()
        self.view = TaskItemView()
        self.data = data
        self.downloadModelList = []
        self.path = ''

        if isinstance(data, Album):
            self.__initAlbum__(data)
        elif isinstance(data, Track):
            self.__initTrack__(data)
        elif isinstance(data, Video):
            self.__initVideo__(data)
        elif isinstance(data, Playlist):
            self.__initPlaylist__(data)

        self.view.connectButton('retry', self.__btnFuncRetry__)
        self.view.connectButton('cancel', self.__btnFuncCancel__)
        self.view.connectButton('delete', self.__btnFuncDelete__)
        self.view.connectButton('open', self.__btnFuncOpen__)

        self.SIGNAL_REFRESH_VIEW.connect(self.__refresh__)

    def __refresh__(self, stype: str, text: str):
        if stype == "setPic":
            self.view.setPic(text)
        elif stype == "addListItem":
            for index, item in enumerate(text):
                downItem = DownloadItemModel(index + 1, item, self.path)
                self.view.addListItem(downItem.view)
                self.downloadModelList.append(downItem)

    def __btnFuncRetry__(self):
        for item in self.downloadModelList:
            item.retry()

    def __btnFuncCancel__(self):
        for item in self.downloadModelList:
            item.stopDownload()

    def __btnFuncDelete__(self):
        for item in self.downloadModelList:
            item.stopDownload()

    def __btnFuncOpen__(self):
        if self.path == '':
            return
        if os.path.exists(self.path):
            os.startfile(self.path)

    def __initAlbum__(self, data: Album):
        self.path = tidalImp.getBasePath(data)
        
        title = data.title
        desc = f"by {tidalImp.getArtistsNames(data.artists)} " \
               f"{tidalImp.getDurationString(data.duration)} " \
               f"Track-{data.numberOfTracks} " \
               f"Video-{data.numberOfVideos}"
        self.view.setLabel(title, desc)

        def __thread_func__(model: TaskItemModel, album: Album):
            cover = tidalImp.getCoverData(album.cover, '1280', '1280')
            model.SIGNAL_REFRESH_VIEW.emit('setPic', cover)

            msg, tracks, videos = tidalImp.getItems(album.id, Type.Album)
            if not aigpy.stringHelper.isNull(msg):
                model.view.setErrmsg(msg)
                return

            model.SIGNAL_REFRESH_VIEW.emit('addListItem', tracks + videos)
            print('__initAlbum__')
            time.sleep(1)

        _thread.start_new_thread(__thread_func__, (self, data))

    def __initTrack__(self, data: Track):
        pass

    def __initVideo__(self, data: Video):
        pass

    def __initPlaylist__(self, data: Playlist):
        pass
